from argparse import ArgumentParser
from pathlib import Path
import logging
import os
import sys
import io
from typing import List

from testsuite.toolbox.scheduling import JobList
from testsuite.toolbox.api import OllamaAdapter


def main() -> None:
    # Argumente verarbeiten
    parser = ArgumentParser(prog="LLM Stegatest")
    # Optionen
    parser.add_argument(
        "-v",
        action="count",
        dest="verbosity",
        default=0,
        help="Verbositätslevel: "
        + "0 (standard): nur Fehler werden ausgegeben; 1: +Warnungen; 2: +Informationen; 3: Debug-Infos",
    )
    parser.add_argument(
        "-o",
        dest="output_path",
        type=str,
        default=None,
        help="Pfad, auf den Output geschrieben werden soll, falls nicht angegeben wird der Output einfach auf der Konsole ausgegeben",
    )
    parser.add_argument(
        "-l",
        dest="logging_path",
        type=str,
        default=None,
        help="Pfad, auf den Logs geschrieben werden sollen, falls nicht angegeben wird auf der Konsole (STDOUT) geloggt.",
    )
    parser.add_argument(
        "--latex",
        dest="latex",
        action="store_true",
        help="Falls verwendet, wird der Output stattdessen in einem Format erzeugt, der in Latex-Dokumenten genutzt werden kann.",
    )
    parser.add_argument(
        "--show-think",
        dest="think",
        action="store_true",
        help="Schneide den Output innerhalb der Thinking Tags bei Modellen wir Qwen 2.5 14B R1 nicht aus dem Output heraus."
        + "Kann ggf. verfälschte Ergebnisse bei Benchmarks erzeugen!!",
    )
    # Argumente
    parser.add_argument(
        "model", type=str, help="Model, für das Testsuite ausgeführt werden so"
    )
    parser.add_argument(
        "prompts",
        type=str,
        default=None,
        help="Prompts, die verwendet werden sollen (Standard: suche in ./prompts nach toml Dateien)",
        nargs="*",
    )
    args = parser.parse_args()

    # Setup für Logging
    # Hack, um sicherzustellen, dass gewählter Pfad writeable durch ausführenden Nutzer ist
    logging_kwargs = {
        "format": "[%(levelname)s] %(message)s",
        "level": max(40 - args.verbosity * 10, 10),
    }

    if args.logging_path is not None:
        logpath: Path = Path(args.logging_path)
        # überprüfe, ob Log geschrieben werden kann
        if not os.access(logpath.absolute(), os.W_OK):
            raise Exception(f"Kann nicht auf {logpath} schreiben!")

        logging_kwargs["filename"] = logpath.absolute()
    logging.basicConfig(**logging_kwargs)

    # Variablenbezeichner aus args mappen (damit statische Analyse die finden kann)
    modelname: str = args.model

    # Hier gehts los!
    # Zu verwendender Adapter, der Prompts in Calls an das Model umwandelt
    adapter = OllamaAdapter(modelname=modelname, cut_thinking_tags=not args.think)

    # Pfade der Prompts
    paths: List[Path] = []
    if len(args.prompts) > 0:
        # übersetze strings in Pfad und überprüfe, ob die angegebenen Dateien existieren
        for p in args.prompts:
            path = Path(p)
            if path.exists():
                paths.append(path)
            else:
                logging.info(f"{p} wurde nicht gefunden, überspringe")
    else:
        base = Path("./prompts")
        if not base.exists() and not base.is_dir():
            logging.critical(
                "Keine Prompts angegeben und kann prompts-Ordner nicht finden, beende"
            )
            exit(1)
        for file in base.iterdir():
            if file.match("*.toml"):
                logging.info(f"Verwende {file.name}")
                paths.append(file)
        if len(paths) == 0:
            logging.critical("Konnte keine toml-Dateien im Ordner prompts finden")

    joblist: JobList = JobList(paths=paths)
    results = joblist.execute_on(adapter)

    # Überprüfe, ob Outputpfad schreibbar ist, falls angegeben
    if args.output_path is not None:
        outpath: Path = Path(args.output_path)
        # überprüfe, ob Log geschrieben werden kann
        if not os.access(outpath.parent, os.W_OK):
            raise Exception(f"Kann nicht auf {outpath} schreiben!")

        outfile: io.TextIOWrapper = open(outpath.absolute(), "w")
    else:
        # wir ignorieren ausnahmsweise den Typ, weil auf stdout write: str -> IO definiert ist
        outfile = sys.stdout  # type: ignore[assignment]

    for res in results:
        if args.latex:
            outfile.write(res.str_latex() + "\n\n")  # Doppelabsatz weil LaTeX
        else:
            outfile.write(str(res) + "\n")


if __name__ == "__main__":
    main()
