"""
Verbindung zu Ollama Server, auf dem entsprechendes Modell ausgeführt werden kann
"""

import logging
import ollama

# from llama_cpp import Llama
from typing import List, Dict

from testsuite.toolbox.prompt import Prompt


class PromtCutter:
    @staticmethod
    def cut(text, keywords):
        """
        :param text: Text
        :param keywords: Liste von Schlüsselwörtern
        :return: Gekürzter String-Text
        """
        cut_text = ""
        words = text.split()
        highlight = False

        for word in words:
            if word in keywords:
                highlight = not highlight
                continue  # Skip the keyword itself
            if not highlight:
                cut_text += word + " "

        return cut_text.strip()

    @staticmethod
    def highlight_keywords(text, keywords):
        """
        :param text: Text, der analysiert werden soll
        :param keywords: Liste von Schlüsselwörtern
        :return: Text mit hervorgehobenen Schlüsselwörtern
        """
        highlighted_text = ""
        words = text.split()
        highlight = False

        for word in words:
            if word in keywords:
                highlight = not highlight
                highlighted_text += f"\033[1;31m{word}\033[0m "  # highlight Red
            elif highlight:
                highlighted_text += f"\033[1;34m{word}\033[0m "  # highlight Blue
            else:
                highlighted_text += word + " "

        return highlighted_text.strip()


class ModelAdapter:
    """
    generische Klasse für Adapter
    """

    @classmethod
    def _prep_dicts(cls, prompt: Prompt) -> List[Dict]:
        userprompt_dicts = [{"role": "user", "content": e} for e in prompt.user_prompt]
        return [
            {"role": "system", "content": str(prompt.system_prompt)}
        ] + userprompt_dicts

    def __init__(self, modelname: str, cut_thinking_tags: bool = False) -> None:
        raise NotImplementedError()

    def execute(self, prompt: Prompt) -> str:
        raise NotImplementedError()


class OllamaAdapter(ModelAdapter):
    def __init__(self, modelname: str, cut_thinking_tags: bool = False) -> None:
        """
        Initialisiert die Ausführung von Prompts auf Models. Falls ein Model
        nicht existiert, wird hier ein Fehler erzeugt.
        """
        try:
            ollama.show(modelname)
        except ollama.ResponseError as e:
            if e.status_code == 404:
                logging.critical(
                    f"Modell {modelname} existiert nicht, überprüfe den angegebenen Namen oder"
                    + " installiere das Model zuerst!",
                )
                exit(1)
        except ConnectionError:
            logging.critical(
                "Keine Verbindung zu Ollama Server, starte den Ollama Serverprozess mit 'ollama serve'",
            )
            exit(1)
        logging.info(f"Ollama läuft und {modelname} wurde gefunden!")

        # Alles Paletti, das Modell kann genutzt werden
        self.modelname = modelname
        self.thinking_tags = cut_thinking_tags

    def execute(self, prompt: Prompt) -> str:
        messages = self._prep_dicts(prompt)

        response: ollama.ChatResponse = ollama.chat(
            model=self.modelname,
            messages=messages,
        )
        content = response["message"]["content"]

        if self.thinking_tags:
            content = PromtCutter.cut(content, ["<think>", "</think>"])

        return content
