"""
Repräsentation eines Prompts, ist außerhalb von scheduling.py um circular imports zu vermeiden.

- Prompt: Eine Kombination von System und Nutzerinputs, die ein Model bekommen kann, sowie Informationen
  über das erwartete Verhalten des Models (Capability)
"""

from typing import List


class Prompt:
    def __init__(
        self, testname: str, system: str | None, user: str | None, expected: str
    ):
        self.name = testname
        self.system_prompt: str | None = system
        self.user_prompt: List[str] = [user] if user is not None else []
        self.expected_behaviour = expected

    def __str__(self):
        return (
            f"Name: {self.name} | System: {self.system_prompt} "
            + f"| User: {str(self.user_prompt)} | Expected: {self.expected_behaviour}"
        )

    def combine(self, other: "Prompt", merge_names: bool = False) -> None:
        """
        Füge weitere Nutzerprompts an diesen Prompt an, schreibt auf das Objekt,
        also müssen entsprechende Daten vor dem Zusammenfügen kopiert werden!

        Der Testname und das erwartete Verhalten von OTHER überschreibt den des
        Objektes, auf dem die Methode ausgeführt wurde. Simulierter Nutzerinput
        wird nacheinander ausgeführt (analog der Vorgehensweise im Paper) und
        Systemeingaben werden einfach konkateniert.

        Bsp. promptSetup ist input von promptTestCapa1:
        ```py
        promptTestCapa1Full = promptSetup.copy()
        promptTestCapa1Full.combine(promptTestCapa1)
        ```
        """
        self.name = other.name if not merge_names else self.name + other.name
        self.user_prompt = self.user_prompt + other.user_prompt
        self.expected_behaviour = other.expected_behaviour
        # kombiniere beide Systemprompts mit \n, aber nur falls beide existieren.
        if self.system_prompt is None:
            self.system_prompt = other.system_prompt
        else:
            if other.system_prompt is not None:
                self.system_prompt = self.system_prompt + "\n" + other.system_prompt
            # else keine Änderung notwendig
