"""
Alle Klassen, die die Ausführung der Prompts planen

Wieso auch immer ich hier einen Job-Scheduler gebaut habe weiß ich nicht, aber wird schon passen!!!

- Job: Ein Job oder eine Kombination von Jobs, die jeweils ausgeführt werden sollten, um eine Capability
  zu überprüfen
- JobList: Liste aller auszuführenden Jobs mit Zwischenspeicher der vorherigen Job-Outputs
- JobResult: Das Ergebnis eines Jobs, wird am Ende als Ergebnis der Ausführung der Jobliste erzeugt
"""

from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict
import logging
import tomllib

from tabulate import tabulate

from testsuite.toolbox.api import ModelAdapter
from testsuite.toolbox.prompt import Prompt


@dataclass
class JobResult:
    name: str
    expected: str
    output: str

    system_prompt: str | None
    user_prompt: List[str] | None

    def _make_str(
        self, tablefmt: str = "rounded_grid", limit_widths: bool = True
    ) -> str:
        line_width = 80
        # Erstelle Tabelle mit Tabulate
        data = [
            ["Testname", self.name],
            ["Erwartung", self.expected],
        ]

        # if self.system_prompt is not None:
        #     data.append(["Systemprompt", self.system_prompt])
        #
        # if self.user_prompt is not None:
        #     for p in self.user_prompt:
        #         data.append(["Nutzeranfrage", p])
        if self.user_prompt is not None:
            data.append(["Nutzeranfrage", self.user_prompt[-1]])

        data.append(["Output", self.output])

        return tabulate(
            data,
            headers="firstrow",
            tablefmt=tablefmt,
            maxcolwidths=[line_width, line_width] if limit_widths else [None, None],
        )

    def __str__(self):
        return self._make_str()

    def str_latex(self):
        table = self._make_str(tablefmt="latex", limit_widths=False)
        table = table.replace("{ll}", "{p{5cm}|p{9cm}}")
        return table


class Job:
    """
    Klasse, die einen Model-Job darstellt
    """

    name: str
    requires: str | None

    def __init__(
        self, prompt: Prompt, requires: str | None = None, name: str | None = None
    ):
        """
        Input: Zu verwendender Prompt und der Name des Jobs, dessen Ergebnis zur Ausführung dieses Jobs
        benötigt wird, falls kein Job benötigt: None
        """
        self.name = prompt.name if name is None else name
        self.requires = requires
        self._prompt = prompt

    def execute_on(
        self, adapter: ModelAdapter, prev_result: JobResult | None = None
    ) -> JobResult:
        """
        Führt Prompt auf Adapter aus. Falls das Ergebnis eines vorherigen Jobs verwendet werden muss,
        wird das Result hier verwendet.

        Input: adapter, vorheriges JobResult
        """
        if prev_result is not None:
            # ersetze {out} in jedem Nutzerprompt mit dem vorherigen Output
            self._prompt.user_prompt[:] = map(
                lambda x: str.format_map(x, {"out": prev_result.output}),
                self._prompt.user_prompt,
            )

        # damit kann der Prompt nun im Adapter verwendet werden
        output = adapter.execute(self._prompt)

        return JobResult(
            name=self.name,
            expected=self._prompt.expected_behaviour,
            output=output,
            system_prompt=self._prompt.system_prompt,
            user_prompt=self._prompt.user_prompt,
        )


class JobList:

    jobs: List[Job]

    def __init__(self, paths: List[Path]):
        """
        Lade Jobs aus prompts.toml Datei. Erzeugt eine Liste aller Jobketten, die
        am Ende von einem Model ausgeführt werden sollen.
        """
        prompts: Dict = {}
        for pfile in paths:
            with pfile.open("rb") as file:
                prompts = prompts | tomllib.load(file)

        prompt_objects: Dict[str, Prompt] = {}
        job_objects: Dict[str, Job] = {}

        """
        Fahrplan: Wir wollen uns eine Liste an abzuarbeitenden Aufgaben erstellen, also
        eine Liste an Prompt Objekten. Da wir Abhängigkeiten der Objekte untereinander möglich
        machen wollen, müssen die "rohen Daten" zwischengespeichert werden, da am Ende nur eine
        Untermenge aller Prompts ausgeführt werden soll (ebenjene, die exec=true gesetzt
        haben).

        Prompts können kombiniert werden, bis wir an eine Stelle kommen, bei der der Input eines
        vorherigen Prompts verwendet wird. Hier muss ein neuer Chat gestartet werden, da wir
        derartige Aktionen nur ausführen, wenn Die Übertragung von Nachrichten simuliert werden
        soll.

        Damit am Ende nicht der gleiche Prompt 10 Mal mit der Intention das gleiche Ergebnis zu
        haben ins Model geschmissen wird, werden Ergebnisse eines Jobs in der Ausführung der
        Jobliste zwischengespeichert.
        """
        for pname, spec in prompts.items():
            # Prompts, die in diesem Schritt erstellt werden
            new_in_this_step: List[Prompt] = []
            # Basisprompt in diesem Schritt
            new_prompt: Prompt = Prompt(
                testname=pname,
                system=spec.get("system", None),
                user=spec.get("user", None),
                expected=spec["expected"],
            )

            # Merge mit vorherigem Prompt, falls input gegeben
            # und previous_out_as_input = false
            job_used_as_input: str | None = spec.get("input", None)
            parent_pobject: str | None = spec.get("parent", None)
            should_be_executed = spec["exec"]

            if parent_pobject is not None:
                if isinstance(parent_pobject, str):
                    # merge mit Eltern-Prompt falls String
                    if parent_pobject not in prompt_objects.keys():
                        logging.critical(
                            f"Parsingfehler: Input {parent_pobject} für {pname} ist noch nicht definiert worden"
                        )
                        exit(1)

                    merged = deepcopy(prompt_objects[parent_pobject])
                    merged.combine(new_prompt)

                    prompt_objects[pname] = merged
                    # Registrieren
                    new_in_this_step.append(merged)
                    # else: wir müssen den Prompt nicht erweitern sondern einen extra Job erstellen
                elif isinstance(parent_pobject, list):
                    for p in parent_pobject:
                        if p not in prompt_objects.keys():
                            logging.critical(
                                f"Parsingfehler: Input {parent_pobject} für {pname} ist noch nicht definiert worden"
                            )
                            exit(1)
                        merged = deepcopy(prompt_objects[p])
                        merged.combine(new_prompt, merge_names=True)

                        new_in_this_step.append(merged)
                        prompt_objects[merged.name] = merged
                else:
                    logging.critical(f"{pname} Eintrag parent ist weder str not Liste")
            else:
                new_in_this_step.append(new_prompt)
                prompt_objects[pname] = new_prompt

            # Jetzt sind alle Prompt bis zu diesem Eintrag gesammelt!
            # Nun überprüfen wir, ob wir einen Job erstellen müssen...
            if should_be_executed:
                for prompt in new_in_this_step:
                    new_jobs: List[Job] = []
                    if job_used_as_input is not None:
                        if isinstance(job_used_as_input, str):
                            new_jobs.append(
                                Job(
                                    prompt=prompt_objects[prompt.name],
                                    requires=job_used_as_input,
                                )
                            )
                        elif isinstance(job_used_as_input, list):
                            for used_input in job_used_as_input:
                                new_jobs.append(
                                    Job(
                                        # mehrere inputs -> Userprompt wird mit erster Nachricht formatiert
                                        # also muss der Prompt hier per deepcopy kopiert werden, da sonst
                                        # mit Referenzen gearbeitet wird
                                        prompt=deepcopy(prompt_objects[prompt.name]),
                                        requires=used_input,
                                        name=f"{used_input}{prompt.name}",
                                    )
                                )
                    else:
                        new_jobs.append(Job(prompt=prompt_objects[prompt.name]))

                    # Speichere unsere neuen Jobs ab
                    for job in new_jobs:
                        job_objects[job.name] = job

            # Überprüfe, ob für jeden Job mit Input der entsprechende Input-Job existiert
            for jobname, job in job_objects.items():
                if (job.requires is not None) and (
                    job.requires not in job_objects.keys()
                ):
                    logging.critical(
                        f"Von Job {jobname} als Vorbedingung deklarierter Job {job.requires} wurde nicht in die "
                        + "Jobliste aufgenommen. Ist 'exec=true' in der Definition des Inputs?"
                    )
                    exit(1)

        # Ok, los gehts!
        self.jobs = [v for _, v in job_objects.items()]

    def execute_on(self, adapter: ModelAdapter) -> List[JobResult]:
        """
        Führt alle Jobs auf adapter aus.
        """
        cached_results: Dict[str, JobResult] = {}

        jobcount = len(self.jobs)
        logging.info(f"Starte {jobcount} Jobs:")
        for job in self.jobs:
            logging.info(f" - {job.name}")

        for i, job in enumerate(self.jobs):
            if job.requires is not None:
                result = job.execute_on(adapter, cached_results[job.requires])
            else:
                result = job.execute_on(adapter)

            logging.info(f"({i+1:>3}/{jobcount:>3}) Job {job.name} Fertig")
            cached_results[job.name] = result

        return [result for _, result in cached_results.items()]
